#!/usr/bin/perl
# vnumlparser.pl
#
# ---------------------------------------------------------------------------------
# VNUML parser. Version 1.3.1
#
# Author: Fermn Galn Mrquez (galan@dit.upm.es)
# Copyright (C) 2004, 	DIT-UPM
# 			Departamento de Ingenieria de Sistemas Telematicos
#			Universidad Politecnica de Madrid
#			SPAIN
#			
# Available at:		http://www.dit.upm.es/vnuml         (IPv4)
#			http://www.upm.euro6ix.org/vnuml    (IPv6)
#
# ----------------------------------------------------------------------------------
#
# VNUML has been developed by DIT-UPM with the partial support from the European 
# Commission in the context of the Euro6IX research project: 
#
#     http://www.euro6ix.org
#
# ----------------------------------------------------------------------------------
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# An online copy of the licence can be found at http://www.gnu.org/copyleft/gpl.html
# -----------------------------------------------------------------------------------

# Becuase of I'm not a Perl guru (in fact, this program is my first development
# in this lenguage :) is possible that the code wouldn't be too tidy.
# The careful reader will discover many bad habits due to years of C
# programming. How many times had I reinvented the wheel in this code? ;)
#
# Anyway, I really apreciate any comment or suggestion about the
# code (and about the english language :) to galan@dit.upm.es

# A word about SSH!
#
# In order to perfom management task (copy files, execute commands, etc)
# SSH is used from the host to the UMLs. To avoid performing interactive
# authentication procedure (that is, write the password) in each access,
# key configuration is desirable.
#
# To do so, we have to generate the pair of keys in the host. In the case
# of using SSH v1, the command is
#
#    ssh-keygen -t rsa1
#
# This generates a pair of files (in $HOME/.ssh/ by default)
#
#    identity		->	in $HOME/.ssh/ in the host
#    identity.pub	->	append it to $HOME/.ssh/authorized_keys file in the UML
#
# Usually $HOME will be /root
#
# Global tag <ssh_key> can be used pointing to the file identity.pub (absolute
# pathname) in the host, in order to the parser perform automaticly the intallation
# of the key in the authorized_keys of the UMLs.
#
# Anyway, in the first access with SSH to the UMLs, we have to confirm the
# key that authenticates the sshd server (in the UJML).
#
# SSHv2 works in the same way, althought it is not yet implemented. If you
# want to considerate this possibility look at man pages of ssh-keygen(1) and
# ssh(1).

# A word about boot process!
#
# The following is a comment to the -t mode.
#
# The master filesystem has to have a SXXumlboot symlink that points
# to /opt/umlboot, the actual boot script, built by the parser in
# certain cases.
#
# In addition, by default, master filesytem has to have the /etc/fstab
# configured to automactly mount /dev/ubd1 at /opt in its LAST LINE
#
# There are three boot modes, depending of the <filesystem> type option.
#
# a) type="direct"
#    Filesystem in the <filesystem> tag is mounted. /opt/umlboot is build
#    The last line of /etc/fstab is deleted.
#    The UML is booted using the filesystem as ubd0, without ubd1.
#
# b) type "copy"
#    The same as a), but the original filesystem file is not used. Instead, a copy
#    is used (in /var/uml)
#
# c) type "cow"
#    A tiny filesystem is created. This filesystem is mounted, a suitable umlboot
#    file is built. The UML is booted using the filesystem in the <filesystem> tag as
#    ubd0 in COW mode and with the tiny filesystem as ubd1.
#
# To sumarize, the master filesystem have to acomplish with the following requeriments:
#
# - /opt directory (empty)
# - symlink at rc point (/etc/rc.d/rc3.d/S11umlboot is suggested) pointing to
#   /opt/umlboot
# - /root/.ssh is a symlink pointing to /opt
# - /etc/fstab with the following last line:
#	/dev/ubd1      /opt    ext2   defaults 0 0
#
#

###########################################################
# Use clauses

use strict;
use XML::DOM;
use XML::DOM::ValParser;
use File::Basename;
use Getopt::Std;
use IO::Socket;
use Net::IPv6Addr;

###########################################################
# Global variables

# Execution mode
my $exe_mode;				# Execution mode (used by &execute)
	use constant EXE_DEBUG => 0;	#	- does not execute, only shows
	use constant EXE_VERBOSE => 1;	#	- executes and shows
	use constant EXE_NORMAL => 2;	#	- executes

# Version information
my $version = "1.3.1";
my $release = "08/01/2004";
my $branch = "";		# "letter" versions are not stable, previous to stable version

# Devault values for some variables
my $basename = basename $0;
my $mount_point = "/mnt/uml";
my $vnuml_dir = "/var/vnuml";
my $tmp_dir = "/tmp";
my $valid_fail = 0;	          # flag used to detect error during XML validation
my $max_name_length = 7;          # max lengthfor names in <vm> and <net>
my $simname;		          # simulation's name
my %vm_hash;		          # list of machines to which apply -s, -t and -p modes
my $ip_offset = 0;	          # by default, offset 0 for IPv4 management interfaces addresses
my $host_mapping = 0;	          # by default, management addresses are not mapped in /etc/hosts
my $promisc = "promisc";          # by default, interfaces are set up in promiscous mode
my $stp = 0;		          # STP configuration in the bridges. By default is not set
my $automac_offset = 0;	          # automatic MAC addresses asignment
my $shell="/bin/bash";	          # shell used by the script that vnumlparser generates
my $tun_device = "/dev/net/tun";  # default tun device
my $start_time;                   # the moment when the parsers start operation

$exe_mode = EXE_NORMAL;
my $verb_prompt = "host> ";	# default promtp used in versbose and debug modes por defecto (to set to "" if you don't want to use any prompt)
my $delay = "5s";		# reference delay used by vnumparser (mounting operations and blocks)
my $scan_mode = "0";		# mode of UML aliveness detection
my $boot_delay = "0";		# by default, 0 seconds for boot delay

# List of mandatory binaries
my @binaries_mandatory = ("touch", "rm",  "echo", "modprobe", "tunctl", 
   "ifconfig", "brctl", "cp", "dd", "mke2fs", "mount", "cat", "umount", 
   "hostname", "route", "scp", "chmod", "ssh", "uml_mconsole",
   "date", "ps", "grep", "pidof", "kill");

# List of optional binaries for VLAN support
my @binaries_vlan = ("vconfig");

# List of optional binaries for screen support
my @binaries_screen = ("screen");

# Hash to store paths of binaries
my %bp;

# Hash to store kernel names
my %kernel_names;

###########################################################
# To get the invocation arguments

our($opt_t,$opt_s,$opt_p,$opt_r,$opt_d,
    $opt_m,$opt_c,$opt_g,$opt_v,$opt_F,
    $opt_V,$opt_T,$opt_o,$opt_M,$opt_k,
    $opt_i,$opt_B,$opt_S,$opt_w,$opt_e,
    $opt_P);
getopts('t:s:p:r:d:m:c:T:gvFVo:M:kiBS:w:e:P:');

# To check arguments consistency
# 1. To use -t, -s, -p, -d, -V or -P at the same time
my $how_many_args;
$how_many_args++ if ($opt_t);
$how_many_args++ if ($opt_s);
$how_many_args++ if ($opt_p);
$how_many_args++ if ($opt_r);
$how_many_args++ if ($opt_d);
$how_many_args++ if ($opt_P);
$how_many_args++ if ($opt_V);
if ($how_many_args gt 1) {
   &usage;
   &smartdie ("only one of the following modes at a time: -t, -s, -p, -r, -d, -P or -V\n");
}
if ($how_many_args lt 1) {
   &usage;
   &smartdie ("missing -t, -s, -p, -r, -d, -P or -V mode\n");
}
if (($opt_F) && (!$opt_d)) {
   &usage;
   &smartdie ("Option -F only makes sense with -d mode\n");
}
if (($opt_B) && ($opt_F) && ($opt_d)) {
   &smartdie ("Option -F and -B are incompabible\n");
}
if (($opt_o) && (!$opt_t)) {
   &usage;
   &smartdie ("Option -o only makes sense with -t mode\n");
}
if (($opt_w) && (!$opt_t)) {
   &usage;
   &smartdie ("Option -w only makes sense with -t mode\n");
}
if (($opt_e) && (!$opt_t)) {
   &usage;
   &smartdie ("Option -e only makes sense with -t mode\n");
}
if (($opt_M) && !(($opt_s)||($opt_p)||($opt_r))) {
   &usage;
   &smartdie ("Option -M only makes sense with -s, -p or -r modes\n");
}
if (($opt_S) && ($opt_P)) {
   &usage;
   &smartdie ("Option -S does not make sense in -P mode\n");
}
# Version mode
if ($opt_V) {
   print "This is $basename version: $version" . "$branch ($release)\n";
   print "Fermn Galn Mrquez. galan\@dit.upm.es\n";
   exit(0);
}

# 2. To check that the source file exits
my $input_file;
$input_file = $opt_t if ($opt_t);
$input_file = $opt_s if ($opt_s);
$input_file = $opt_p if ($opt_p);
$input_file = $opt_r if ($opt_r);
$input_file = $opt_d if ($opt_d);
$input_file = $opt_P if ($opt_P);
if (! -f $input_file) {
   &smartdie ("file $input_file is not valid (perhaps does not exists)\n");
}

# 3. Optional arguments
$exe_mode = EXE_VERBOSE if ($opt_v);
$exe_mode = EXE_DEBUG if ($opt_g);
$mount_point = chompslash($opt_m) if ($opt_m);
$vnuml_dir = chompslash($opt_c) if ($opt_c);
$tmp_dir = chompslash($opt_T) if ($opt_T);
$scan_mode = $opt_S if ($opt_S); 
$boot_delay = $opt_w if ($opt_w);   # Shoul check if $opt_w is a valid positive integer. Check it!

# 4. To check mount point, vnuml_dir and tmp_dir
&smartdie ("mount point $mount_point is not a valid directory\n") if (! -d $mount_point);
&smartdie ("vnuml_dir $vnuml_dir is not a valid directory\n") if (! -d $vnuml_dir);
&smartdie ("tmp_dir $tmp_dir is not a valid directory\n") if (! -d $tmp_dir);

# 5. To check mandatory binaries
if (&check_binaries(@binaries_mandatory) != 0) {
   &smartdie ("some required binary files are missing\n");
}

# 6. To check optional screen binaries, if needed
if (($opt_e) && (&check_binaries(@binaries_screen) != 0)) {
   &smartdie ("screen binary is missing\n");
}

# 7. To check if the DTD file is present
open INPUTFILE, "$input_file";
while (<INPUTFILE>) {
   if (/<!DOCTYPE vnuml SYSTEM "(.*)">/) { 
      if (-f $1) {
         last;
      }
      else {
         &smartdie ("DTD file $1 is not valid (perhaps does not exists)\n");
      }
   }
} 
close INPUTFILE;

# Interactive execution (press a key after each command)
my $exe_interactive = $opt_v & $opt_i;

# To build DOM tree parsing file
my $parser;
if ($opt_k) {
   # Non-validator parser
   $parser = new XML::DOM::Parser;
}
else {
   # Validator parser
   $parser = new XML::DOM::ValParser;
}
my $doc;
eval {
   local $XML::Checker::FAIL = \&validation_fail;
   $doc = $parser->parsefile($input_file);
};
if ($valid_fail) {
   # Validation errors
   &smartdie ("$input_file is not a well-formed VNUML file\n");
}
# Semantic check (in addition to validation)
if (my $err_msg = &check_semantics) {
   &smartdie ("$err_msg\n");
}


###########################################################
# Global data capture

# Simulation's name
my $simname_list = $doc->getElementsByTagName("simulation_name");
if ($simname_list->getLength == 1) {
   $simname = &text_tag($simname_list->item(0));
}
else {
   &smartdie ("simulation name is missing\n");
}

# To check optional binaries for VLAN support (if applicable)
if (&check_tag_attribute("net","vlan") != 0) {
   if (&check_binaries(@binaries_vlan) != 0) {
      &smartdie ("some required binary files for VLAN support are missing\n");
   }
}

# To use <offset>, if present 
my $offset_list = $doc->getElementsByTagName("ip_offset");
if ($offset_list->getLength == 1) {
   $ip_offset = &text_tag($offset_list->item(0));
}

# To use <shell>, if present
my $shell_list = $doc->getElementsByTagName("shell");
if ($shell_list->getLength == 1) {
   $shell = &text_tag($shell_list->item(0));
}

# To use <host_mapping>, if present
my $hostmapping_list = $doc->getElementsByTagName("host_mapping");
if ($hostmapping_list->getLength == 1) {
   $host_mapping = 1;
}

# Network configuration options, if <netconfig> is present
my $netconfig_list = $doc->getElementsByTagName("netconfig");
if ($netconfig_list->getLength == 1) {
   &netconfig($netconfig_list->item(0));
}

# AutoMAC offset, if present
my $automac_list = $doc->getElementsByTagName("automac");
if ($automac_list->getLength == 1) {
   my $att = $automac_list->item(0)->getAttribute("offset");
   if ($att =~ /^$/) {
      $automac_offset = 0;
   }
   else {
      $automac_offset = $att;
   }
}

# Tun device, if <tun_device> is present
my $tun_device = "/dev/net/tun";
my $tun_device_list = $doc->getElementsByTagName("tun_device");
if ($tun_device_list->getLength == 1) {
   $tun_device = &text_tag($tun_device_list->item(0));;
}

# Check tun device existence (as characters device)
&smartdie ("$tun_device is not a valid character device file\n") unless (-c $tun_device);

###########################################################
# Command execution

if ($exe_interactive) {
   print "interactive execution is on: pulse a key after each command\n";
}

# Lock management
if (-f "$vnuml_dir/LOCK") {
   die ("$vnuml_dir/LOCK exists: another instance of $basename seems to be in execution\nIf you are sure that this cann't be happeing in your system, do 'rm $vnuml_dir/LOCK' and try again\n");
}
else {
   &execute ($bp{"touch"} . " $vnuml_dir/LOCK");
   $start_time = time();
}

# Mode selection
if ($opt_t) {
   if ($exe_mode != EXE_DEBUG) {
      &smartdie ("simulation $simname already created\n") if &simulation_exists($simname);
   }
   &mode_t;
}
elsif ($opt_s) {
   if ($exe_mode != EXE_DEBUG) {
     &smartdie ("simulation $simname does not exists: create it with -t before\n")
        unless &simulation_exists($simname);
   }
   &make_vm_hash;
   &mode_sp("start");
}
elsif ($opt_p) {
   if ($exe_mode != EXE_DEBUG) {
      &smartdie ("simulation $simname does not exists: create it with -t before\n")
        unless &simulation_exists($simname);
   }
   &make_vm_hash;
   &mode_sp("stop");
}
elsif ($opt_r) {
   if ($exe_mode != EXE_DEBUG) {
      &smartdie ("simulation $simname does not exists: create it with -t before\n")
        unless &simulation_exists($simname);
   }
   &make_vm_hash;
   &mode_sp("stop");
   &mode_sp("start");
}
elsif ($opt_d) {
   if ($exe_mode != EXE_DEBUG) {
      &smartdie ("simulation $simname does not exists\n")
        unless &simulation_exists($simname);
   }
   &mode_d;
}
elsif ($opt_P) {
   if ($exe_mode != EXE_DEBUG) {
   #   &smartdie ("simulation $simname does not exists\n")
   #     unless &simulation_exists($simname);
   }
   $opt_F = 1;
   &mode_d;		# First, to call destroy mode with force flag activated
   &mode_P;		# Second, purge another things
}
else {
   &smartdie ("if you are seeing this line something terribly horrible has happened...\n");
}

# Remove lock
&execute ($bp{"rm"} . " $vnuml_dir/LOCK");
my $total_time = time() - $start_time;
print "Total time eslapsed: $total_time seconds\n";

exit(0);

#####################################################

# mode_t
#
# Build topology mode
sub mode_t {

   # To make directory to store files related with the topology
   if (! -d "$vnuml_dir/$simname") {
      mkdir "$vnuml_dir/$simname" or &smartdie ("error making directory $vnuml_dir/$simname: $!");
   }

   # To make lock file (it exists while topology is running)
   &execute ($bp{"touch"} . " $vnuml_dir/$simname/lock");

   # If host_mapping is in use, append trailer to /etc/hosts
   if ($host_mapping) {
        my $command = $bp{"date"};
        chomp (my $now = `$command`);
	&execute ($bp{"echo"} . " >> /etc/hosts");
	&execute ($bp{"echo"} . " \\\# virtual machine names added by $basename $version$branch at $now >> /etc/hosts");
   }

   # 1. To create host TUN/TAP devices
   &tun_create;

   # 2. To create bridges (if needed)
   &bridge_create;

   # 3. Host configuration
   &host_config;

   # 4. To link TUN/TAP to the bridges
   &tun_connect;

   # 5. To create boot files for each UML
   &bootfiles;

   # 6. To boot each UML
   &boot;

   # If -B, block until ready
   if ($opt_B) {
      my $time_0 = time();
      while (!&UMLs_ready($scan_mode)) {
         #system($bp{"sleep"} . " $delay");
         sleep($delay);
         my $time_w = time();
         my $interval = $time_w - $time_0;
         print "$interval seconds eslapsed...\n" if ($exe_mode == EXE_VERBOSE);
      }
   }

}

######################################################
# To create TUN/TAP devices

sub tun_create {

   # To start tun module
   !&execute ($bp{"modprobe"} . " tun") or &smartdie ("module tun can not be initialized: $!");

   # To get list of defined <vm> 
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # We get name attribute
      my $name = $vm->getAttribute("name");

      # To create management device (id 0)
      my $tun_if = $name . "-eth0";
      &execute($bp{"tunctl"} . " -t $tun_if -f $tun_device");

      # To get UML's interfaces list
      my $if_list = $vm->getElementsByTagName("if");

      # To process list
      for ( my $j = 0; $j < $if_list->getLength; $j++) {
         my $if = $if_list->item($j);

         # We get id attribute
         my $id = $if->getAttribute("id");

	 # We build TUN device name
	 $tun_if = $name . "-eth" . $id;

         # To create TUN device
	 &execute($bp{"tunctl"} . " -t $tun_if -f $tun_device");

         # To set up device
         &execute($bp{"ifconfig"} . " $tun_if 0.0.0.0 $promisc up");

      }

   }

}


######################################################
# To create bridges

sub bridge_create {

   # To get list of defined <net>
   my $net_list = $doc->getElementsByTagName("net");

   # To process list
   for ( my $i = 0; $i < $net_list->getLength; $i++ ) {
      my $net = $net_list->item($i);

      # We get name attribute
      my $name = $net->getAttribute("name");

      # If bridged does not exists, we create and set up it
      unless (&vnet_exists($name)) {
         &execute($bp{"brctl"} . " addbr $name");
	 if ($stp) {
            &execute($bp{"brctl"} . " stp $name on");
	 }
	 else {
            &execute($bp{"brctl"} . " stp $name off");
	 }
	 &execute($bp{"ifconfig"} . " $name 0.0.0.0 $promisc up");
      }

      # Is there an external interface associated with the bridge?
      my $external_if = $net->getAttribute("external");
      unless ($external_if =~ /^$/) {
         # If there is an external interface associate, to check if VLAN is being used
	 my $vlan = $net->getAttribute("vlan");
	 unless ($vlan =~ /^$/ ) {
	    # If there is not any configured VLAN at this interface, we have to enable it
	    unless (&check_vlan($external_if,"*")) {
               &execute($bp{"ifconfig"} . " $external_if 0.0.0.0 $promisc up");
	    }
	    # If VLAN is already configured at this interface, we haven't to configure it
	    unless (&check_vlan($external_if,$vlan)) {
	       &execute($bp{"modprobe"} . " 8021q");
	       &execute($bp{"vconfig"} . " add $external_if $vlan");
	    }
	    $external_if .= ".$vlan";
	 }
	 # If the interface is already added to the briged, we haven't to add it
	 my @if_list = &vnet_ifs($name);
	 $_ = "@if_list";
	 unless (/\b($external_if)\b/) {
            &execute($bp{"ifconfig"} . " $external_if 0.0.0.0 $promisc up");
	    &execute($bp{"brctl"} . " addif $name $external_if");
	 }
	 # We increase interface use counter
	 &inc_if($external_if);
      }

   }

}


######################################################
# To link TUN/TAP to the bridges

sub tun_connect {

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # We get name attribute
      my $name = $vm->getAttribute("name");

      # To get UML's interfaces list
      my $if_list = $vm->getElementsByTagName("if");

      # To process list
      for ( my $j = 0; $j < $if_list->getLength; $j++) {
         my $if = $if_list->item($j);

         # To get id attribute
         my $id = $if->getAttribute("id");

         # We get net attribute
         my $net = $if->getAttribute("net");
	 my $net_if = $name . "-eth" . $id;

         # We link TUN/TAP device
	 &execute($bp{"brctl"} . " addif $net $net_if");

      }
   }

}


######################################################
# To create boot files for each UML

sub bootfiles {

   # UMLs counter (used to generate IPv4 management addresses)
   my $counter = 0;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # We get name attribute
      my $name = $vm->getAttribute("name");

      # To get filesystem and type
      my $filesystem = &text_tag($vm->getElementsByTagName("filesystem")->item(0));
      my $filesystem_type = $vm->getElementsByTagName("filesystem")->item(0)->getAttribute("type");

      # Copy needed?
      if ($filesystem_type eq "copy") {
         &execute($bp{"cp"} . " $filesystem $vnuml_dir/$simname/$name" . "_fs");
	 $filesystem = "$vnuml_dir/$simname/$name" . "_fs";
      }
      # Auxiliar filesystem (COW) needed?
      if ($filesystem_type eq "cow") {
         &execute($bp{"dd"} . " if=/dev/zero of=$vnuml_dir/$simname/$name" . "_opt_fs count=1 bs=1M");
	 &execute($bp{"mke2fs"} . " -F $vnuml_dir/$simname/$name" . "_opt_fs");
	 $filesystem = "$vnuml_dir/$simname/$name" . "_opt_fs";
      }

      # Whether we are mounting main filesystem or the auxiliar one (when COW is in use),
      # /opt/ or / is used as path
      my $path;
      if ($filesystem_type eq "cow") {
         $path = "/";
      }
      else {
         $path = "/opt/";
      }

      # To mount filesystem (in loop mode)
      &execute($bp{"mount"} . " $filesystem $mount_point -o loop");

      # To install public ssh key in the UML
      my $key_list = $doc->getElementsByTagName("ssh_key");
      # If tag present, add the key
      if ($key_list->getLength == 1) {
         my $keyfile = &text_tag($key_list->item(0));
	 &execute($bp{"cat"} . " $keyfile >> $mount_point$path" . "authorized_keys");
      }

      # To make configuration file
      &UML_bootfile($path, $vm, $counter);

      # To umount filesystem
      # pause is needed or umount may crash (sync problems?)
      &execute($bp{"umount"} . " $mount_point");
      #&execute($bp{"sleep"} . " $delay");
      sleep($delay) unless ($exe_mode == EXE_DEBUG);

      $counter++;
   }

}

sub UML_bootfile {

    my $path = shift;
    my $vm = shift;
    my $number = shift;

    my $vm_name = $vm->getAttribute("name");

    # We open boot file, taking S$boot_prio and $runlevel
    open CONFILE, ">$mount_point$path" . "umlboot"
       or &smartdie ("can not open $mount_point$path" . "umlboot: $!")
       unless ($exe_mode == EXE_DEBUG);
    my $verb_prompt_bk = $verb_prompt;
    $verb_prompt = "$vm_name> ";

    # We begin boot script
    my $command = $bp{"date"};
    chomp (my $now = `$command`);
    &execute("#!$shell",*CONFILE);
    &execute("# UML boot file generated by $basename at $now" ,*CONFILE);

    # 1. To set hostname
    &execute("hostname $vm_name",*CONFILE);

    # 2. Management interface configuration (UML side)
    my ($net, $mask, $host1, $host2) = get_admin_address($number);
    &execute ("ifconfig eth0 $host2 netmask $mask up",*CONFILE);

    # If host_mapping is in use, append trailer to /etc/hosts
    if ($host_mapping) {
        $verb_prompt = $verb_prompt_bk;
	&execute ($bp{"echo"} . " $host2 $vm_name >> /etc/hosts");
        $verb_prompt = "$vm_name> ";
    }

    # 3. Interface configuration

    # To get UML's interfaces list
    my $if_list = $vm->getElementsByTagName("if");

    # To process list
    for ( my $i = 0; $i < $if_list->getLength; $i++) {
       my $if = $if_list->item($i);

       # We get id attribute
       my $id = $if->getAttribute("id");
       my $if_name = "eth" . $id;

       # To set up interface (without address, at the begining)
       &execute ("ifconfig $if_name 0.0.0.0 $promisc up",*CONFILE);

       # 4a. To process interface IPv4 addresses
       # The first address have to be assigned without "add" to avoid creating subinterfaces
       my $ipv4_list = $if->getElementsByTagName("ipv4");
       my $command = "";
       for ( my $j = 0; $j < $ipv4_list->getLength; $j++) {
	  my $ip = &text_tag($ipv4_list->item($j));
	  my $ipv4_mask = $ipv4_list->item($j)->getAttribute("mask");
	  if ($ipv4_mask =~ /^$/) {
	     $ipv4_mask = "255.255.255.0";
	  }
          &execute("ifconfig $if_name $command $ip netmask $ipv4_mask",*CONFILE);
	  if ($command =~ /^$/) {
             $command = "add";
	  }
       }

       # 4b. To process interface IPv6 addresses
       my $ipv6_list = $if->getElementsByTagName("ipv6");
       for ( my $j = 0; $j < $ipv6_list->getLength; $j++) {
	  my $ip = &text_tag($ipv6_list->item($j));
          &execute("ifconfig $if_name add $ip",*CONFILE);
       }

    }

    # 5. Route configuration
    my $route_list = $vm->getElementsByTagName("route");
    for ( my $i = 0; $i < $route_list->getLength; $i++) {
       # Text access is a mess... but it works :)
       my $route_dest = &text_tag($route_list->item($i));
       my $route_gw = $route_list->item($i)->getAttribute("gw");
       my $route_inet = $route_list->item($i)->getAttribute("type");
       # Routes for IPv4 (inet)
       if ($route_inet eq "inet") {
          if ($route_dest eq "default") {
             &execute("route -A inet add default gw $route_gw",*CONFILE);
          } 
          else {
             &execute("route -A inet add -net $route_dest gw $route_gw",*CONFILE);
          }
          #&execute("route -A $route_inet add $route_dest gw $route_gw",*CONFILE);
       }
       # Routes for IPv6 (inet6)
       else {
          if ($route_dest eq "default") {
             &execute("route -A inet6 add 2000::/3 gw $route_gw",*CONFILE);
          }
          else {
             &execute("route -A inet6 add $route_dest gw $route_gw",*CONFILE);
          }
       }
    }

    # 6. Forwarding configuration
    my $forwarding = $vm->getElementsByTagName("forwarding");
    if ($forwarding->getLength == 1) {
       my $f_type = $forwarding->item(0)->getAttribute("type");
       $f_type = "ip" if ($f_type =~ /^$/);
       &execute("echo 1 > /proc/sys/net/ipv4/ip_forward",*CONFILE) if ($f_type eq "ip" or $f_type eq "ipv4");
       &execute("echo 1 > /proc/sys/net/ipv6/conf/all/forwarding",*CONFILE) if ($f_type eq "ip" or $f_type eq "ipv6");
    }

    # Close file and restore prompting method
    $verb_prompt = $verb_prompt_bk;
    close CONFILE unless ($exe_mode == EXE_DEBUG);

    # Management interface configuration (host side)
    &execute ($bp{"ifconfig"} . " $vm_name-eth0 $host1 netmask $mask up");

    # Boot file will be executable
    &execute($bp{"chmod"} . " a+x $mount_point$path". "umlboot");

}


#####################################################
# Host configuration

sub host_config {

   # If host tag is not present, there is nothing to do
   return if ($doc->getElementsByTagName("host")->getLength eq 0);

   my $host = $doc->getElementsByTagName("host")->item(0);

   # To get host's interfaces list
   my $if_list = $host->getElementsByTagName("hostif");

   # To process list
   for ( my $i = 0; $i < $if_list->getLength; $i++) {
      my $if = $if_list->item($i);

      # To get name attribute
      my $net = $if->getAttribute("net");

      # Interface configuration
      # 1a. To process interface IPv4 addresses
      # The first address have to be assigned without "add" to avoid creating subinterfaces
      my $ipv4_list = $if->getElementsByTagName("ipv4");
      my $command = "";
      for ( my $j = 0; $j < $ipv4_list->getLength; $j++) {
         my $ip = &text_tag($ipv4_list->item($j));
         my $ipv4_mask = $ipv4_list->item($j)->getAttribute("mask");
         if ($ipv4_mask =~ /^$/) {
	     $ipv4_mask = "255.255.255.0";
	 }
         &execute($bp{"ifconfig"} . " $net $command $ip netmask $ipv4_mask");
	 if ($command =~ /^$/) {
             $command = "add";
	 }
      }

      # 2a. To process interface IPv6 addresses
      my $ipv6_list = $if->getElementsByTagName("ipv6");
      for ( my $j = 0; $j < $ipv6_list->getLength; $j++) {
         my $ip = &text_tag($ipv6_list->item($j));
	 &execute($bp{"ifconfig"} . " $net add $ip");
      }

   }

   # To get host's routes list
   my $route_list = $host->getElementsByTagName("route");
   for ( my $i = 0; $i < $route_list->getLength; $i++) {
       my $route_dest = &text_tag($route_list->item($i));;
       my $route_gw = $route_list->item($i)->getAttribute("gw");
       my $route_inet = $route_list->item($i)->getAttribute("type");
       # Routes for IPv4 (inet)
       if ($route_inet eq "inet") {
          if ($route_dest eq "default") {
             &execute($bp{"route"} . " -A inet add default gw $route_gw");
          } 
          else {
             &execute($bp{"route"} . " -A inet add -net $route_dest gw $route_gw");
          }
          #&execute($bp{"route"} . " -A $route_inet add $route_dest gw $route_gw");
       }
       # Routes for IPv6 (inet6)
       else {
          if ($route_dest eq "default") {
             &execute($bp{"route"} . " -A inet6 add 2000::/3 gw $route_gw");
          }
          else {
             &execute($bp{"route"} . " -A inet6 add $route_dest gw $route_gw");
          }
       }
    }

    # Enable host forwarding
    my $forwarding = $host->getElementsByTagName("forwarding");
    if ($forwarding->getLength == 1) {
       my $f_type = $forwarding->item(0)->getAttribute("type");
       $f_type = "ip" if ($f_type =~ /^$/);
       &execute($bp{"echo"} . " 1 > /proc/sys/net/ipv4/ip_forward") if ($f_type eq "ip" or $f_type eq "ipv4");
       &execute($bp{"echo"} . " 1 > /proc/sys/net/ipv6/conf/all/forwarding") if ($f_type eq "ip" or $f_type eq "ipv6");
    }

}


######################################################
# To boot each UML

sub boot {

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # If defined screen configuration file, open it
   if (($opt_e) && ($exe_mode != EXE_DEBUG)) {
      open SCREEN_CONF, ">$opt_e"
         or &smartdie ("can not open $opt_e: $!")
   }

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # We get name attribute
      my $name = $vm->getAttribute("name");

      # To get filesystem and type
      my $filesystem = &text_tag($vm->getElementsByTagName("filesystem")->item(0));
      my $filesystem_type = $vm->getElementsByTagName("filesystem")->item(0)->getAttribute("type");

      # If copy type, we haven't to use the tag value
      if ($filesystem_type eq "copy") {
         $filesystem = "$vnuml_dir/$simname/$name" . "_fs";
      }
      # If cow type, we have to check whether particular filesystem exists
      # to set the right boot filesystem. In addition, we set ubd1 
      my $ubd1 = "";
      if ($filesystem_type eq "cow") {
         if (-f "$vnuml_dir/$simname/$name" . "_cow_fs") {
	    $filesystem = "$vnuml_dir/$simname/$name" . "_cow_fs";
	 }
	 else {
            $filesystem = "$vnuml_dir/$simname/$name" . "_cow_fs,$filesystem"
	 }
	 $ubd1 = " ubd1=$vnuml_dir/$simname/$name" . "_opt_fs";
      }

      # UML kernel to being booted
      my $kernel = "/usr/local/bin/linux";
      my $kernel_list = $vm->getElementsByTagName("kernel");
      # If tag present, get the kernel name
      if ($kernel_list->getLength == 1) {
         $kernel = &text_tag($kernel_list->item(0));
      }

      # Boot command line
      my $boot_line = "$kernel ubd0=$filesystem$ubd1 root=/dev/ubd/0";

      # Memory assignment
      my $mem_list = $vm->getElementsByTagName("mem");
      # If <mem> tag is not present, no additional switch is added.
      # In this case, UML boots with a default memory size of 32 Mbytes
      if ($mem_list->getLength == 1) {
            $boot_line .= " mem=" . &text_tag($mem_list->item(0));
      }

      # Management interface
      $boot_line .= " eth0=tuntap,$name-eth0";

      # To process all interfaces

      # To get UML's interfaces list
      my $if_list = $vm->getElementsByTagName("if");

      # To process list
      for ( my $j = 0; $j < $if_list->getLength; $j++) {
         my $if = $if_list->item($j);

         # To get id attribute
         my $id = $if->getAttribute("id");

         # To get MAC address
	 my $mac_list = $if->getElementsByTagName("mac");
         my $mac;
         # If <mac> is not present, we ask for an automatic one (if
	 # <automac> is not enable may be null; in this case UML 
         # autoconfiguration based in IP address of the interface 
         # is used -but it doesn't work with IPv6!)
         if ($mac_list->getLength == 1) {
            $mac = "," . &text_tag($mac_list->item(0));
         }
         else {
	    $mac = &automac($i+1, $id);
            # En la versin 1.0.0 era esto:
	    # $mac = "";
         }

         # To add interface to command line
         $boot_line .= " eth$id=tuntap,$name-eth$id$mac";

      }

      # Background UML execution without consoles by default
      $boot_line .= " uml_dir=$vnuml_dir/$simname/ umid=$name con=null";

      # Process <boot> tag (if present)
      my $boot_list = $vm->getElementsByTagName("boot");
      # If tag present, go into
      if ($boot_list->getLength == 1) {
         my $boot = $boot_list->item(0);

         # Proccess <con0> tag (if present)
         my $con0_list = $boot->getElementsByTagName("con0");
         # If tag present, go into
         if ($con0_list->getLength == 1) {
            my $con0 = &text_tag($con0_list->item(0));
            $boot_line .= " con0=$con0";
            }

      }

      # Boot UML output redirection to a file (if -o was used in parser command line )
      if ($opt_o) {
          if ($opt_o eq "/dev/null") {
             $boot_line .= " &> /dev/null &";
          }
          else {
             $boot_line .= " &> $opt_o.$name &";
          }
      }
      else {
          $boot_line .= " &";
      }

      # Boot command execution
      &execute("$boot_line");
      
      if ($boot_delay != "0") { 
         # Delay needed to avoid weird behaviour with concurrent booting?
         #&execute($bp{"sleep"} . " " . $boot_delay . "s");
         sleep("$boot_delay" . "s") unless ($exe_mode == EXE_DEBUG);
      }

      # Proccess <con0> tag (if present)
      if (($opt_e) && ($exe_mode != EXE_DEBUG)) {
         my $con0_list = $vm->getElementsByTagName("con0");
         # If tag present, go into
         if ($con0_list->getLength == 1) {
            my $con0 = &text_tag($con0_list->item(0));

            # Special treatment only with pts devices
            if ($con0 eq "pts") {
               my $pts = "";
               while ($pts =~ /^$/) {	# I'm sure that this loop could be smarter, but it works :)
                  print "Trying to get con0 pts...\n" if ($exe_mode == EXE_VERBOSE);
                  sleep("1s");		# Needed to avoid  syncronization problems
                  my $command = $bp{"uml_mconsole"};
                  my $mconsole_query = `$command $vnuml_dir/$simname/$name/mconsole config con0 2> /dev/null`;
                  if ($mconsole_query =~ /^OK pts:(.*)$/) {
                     $pts = $1;
                     print "...pts is $pts\n" if ($exe_mode == EXE_VERBOSE);
                  }
               }
               print SCREEN_CONF "screen -t $name $pts\n";
            }
         }
      }

   }

   # Close screen configuration file
   if (($opt_e) && ($exe_mode != EXE_DEBUG)) {
      close SCREEN_CONF;
   }


}

#####################################################

# mode_sp
#
# start simulation and stop simulation modes.
# Both are very similar
sub mode_sp {

   # Mode selection (start or stop).
   my $mode = shift;

   # If -B, block until ready
   if ($opt_B) {
      my $time_0 = time();
      while (!&UMLs_ready($scan_mode)) {
         #system($bp{"sleep"} . " $delay");
         sleep($delay);
         my $time_f = time();
         my $interval = $time_f - $time_0;
         print "$interval seconds eslapsed...\n" if ($exe_mode == EXE_VERBOSE);
      }
   }
   else {
      &smartdie ("UMLs are booted but not ready yet. Wait a while and retry...\n") unless &UMLs_ready($scan_mode);
   }

   # 1. To install configuration files subtree
   &conf_files($mode);

   # 2. To build commands files
   &command_files($mode);

   # 3. To load commands file in each UML
   &install_command_files($mode);

   # 4. To execute commands file in each UML
   &exec_command_files($mode);

   # 5. To execute commands file in host
   &exec_command_host($mode);

}

sub conf_files {

   my $tag = shift;

   # UMLs counter (used to generate IPv4 management addresses)
   my $counter = 0;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # We get name attribute
      my $name = $vm->getAttribute("name");

      # We have to process it?
      unless ($vm_hash{$name}) {
        $counter++;
      	next;
      }

      # We look for <filetree> tag
      my $filetree_list = $vm->getElementsByTagName("filetree");
      for ( my $j = 0; $j < $filetree_list->getLength; $j++ ) {
         my $filetree = $filetree_list->item($j);

         # To get host directory (subtree) to install in the UML
	 my $src = &text_tag($filetree);

	 # To get installation point at the UML
	 my $dest = $filetree->getAttribute("root");

	 # To get momment
	 my $when = $filetree->getAttribute("when");

         # Management interface configuration (UML side)
         my ($net, $mask, $host1, $host2) = get_admin_address($counter);

         # To install subtree (only in the right momment)
	 if (($when eq $tag) || ($when eq "always")) {
            &execute($bp{"scp"} . " -q -r -oProtocol=1 $src/* root\@$host2:$dest");
	 }
      }
      $counter++;
   }

}

# This function uses the first argument to know
# whether it has to work with <start> or <stop>
sub command_files {

   my $tag = shift;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # We get name attribute
      my $name = $vm->getAttribute("name");

      # We have to process it?
      unless ($vm_hash{$name}) {
      	next;
      }

      # We open file
      open COMMAND_FILE, ">$tmp_dir/vnuml.$name.$tag"
         or &smartdie ("can not open $tmp_dir/vnuml.$name.$tag: $!")
	 unless ($exe_mode == EXE_DEBUG);;
      my $verb_prompt_bk = $verb_prompt;
      $verb_prompt = "$name> ";
      my $command = $bp{"date"};
      chomp (my $now = `$command`);
      &execute("#!$shell",*COMMAND_FILE);
      &execute("# UML $tag commands file generated by $basename $version$branch at $now",*COMMAND_FILE);

      # To get commands list
      my $command_list = $vm->getElementsByTagName($tag);

      # To process list, dumping commands to file
      for ( my $j = 0; $j < $command_list->getLength; $j++) {
         my $command = $command_list->item($j);

         # To get type attribute
         my $type = $command->getAttribute("type");

         # Case 1. Verbatim type
	 if ($type eq "verbatim") {
	    # Incluimos el comando tal cual
	    &execute(&text_tag($command),*COMMAND_FILE);
	 }
	 # Case 2. File type
         elsif ($type eq "file") {
	    # We open the file and write commands line by line
	    my $include_file = &text_tag($command);
	    open INCLUDE_FILE, "$include_file"
	       or &smartdie ("can not open $include_file: $!");
	    while (<INCLUDE_FILE>) {
	       chomp;
	       &execute($_,*COMMAND_FILE);
	    }
	    close INCLUDE_FILE;
	 }
	 # Other case. Don't do anything (it would be and error in the XML!)

      }

      # We close file and mark it executable
      close COMMAND_FILE unless ($exe_mode == EXE_DEBUG);
      $verb_prompt = $verb_prompt_bk;
      &execute($bp{"chmod"} . " a+x $tmp_dir/vnuml.$name.$tag");
   }

}

sub install_command_files {

   my $tag = shift;

   # UMLs counter (used to generate IPv4 management addresses)
   my $counter = 0;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # Process it?
      unless ($vm_hash{$name}) {
        $counter++;
      	next;
      }

      # Management interface configuration (UML side)
      my ($net, $mask, $host1, $host2) = get_admin_address($counter);

      # We install the file
      &execute($bp{"ssh"} . " -x -1 -l root $host2 rm -f /root/vnuml.$name.$tag &> /dev/null");
      &execute($bp{"scp"} . " -q -oProtocol=1 $tmp_dir/vnuml.$name.$tag root\@$host2:/root");

      $counter++;
   }

}

sub exec_command_files {

   my $tag = shift;

   # UMLs counter (used to generate IPv4 management addresses)
   my $counter = 0;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # Process it?
      unless ($vm_hash{$name}) {
        $counter++;
      	next;
      }

      # Management interface configuration (UML side)
      my ($net, $mask, $host1, $host2) = get_admin_address($counter);

      # We execute the file
      &execute($bp{"ssh"} . " -x -1 -l root $host2 /root/vnuml.$name.$tag");

      $counter++;
   }

}

sub exec_command_host {

   my $tag = shift;

   # If host <host> is not present, there is nothing to do
   return if ($doc->getElementsByTagName("host")->getLength eq 0);

   # To get <host> tag
   my $host = $doc->getElementsByTagName("host")->item(0);

   # To get commands list
   my $command_list = $host->getElementsByTagName($tag);

   # To process list, dumping commands to a file
   for ( my $j = 0; $j < $command_list->getLength; $j++) {
      my $command = $command_list->item($j);

      # To get type attribute
      my $type = $command->getAttribute("type");

      # Case 1. Verbatim type
      if ($type eq "verbatim") {
         # Incluimos el comando tal cual
         &execute(&text_tag($command));
      }
      # Case 2. File type
      elsif ($type eq "file") {
         # We open file and write commands line by line
         my $include_file = &text_tag($command);
         open INCLUDE_FILE, "$include_file"
            or &smartdie ("can not open $include_file: $!");
         while (<INCLUDE_FILE>) {
            chomp;
            &execute($_);
         }
         close INCLUDE_FILE;
      }
      # Other case. Don't do anything (it would be and error in the XML!)

    }

}

#####################################################

# mode_d
#
# Destroy current simulation mode 
sub mode_d {

   # If -B, block until ready
   if ($opt_B) {
      my $time_0 = time();
      while (!&UMLs_ready($scan_mode)) {
         #system($bp{"sleep"} . " $delay");
         sleep($delay);
         my $time_f = time();
         my $interval = $time_f - $time_0;
         print "$interval seconds eslapsed...\n" if ($exe_mode == EXE_VERBOSE);
      }
   }
   else {
      if (!$opt_F) {
         &smartdie ("UMLs are booted but not ready yet. Wait a while or use -F...\n") unless &UMLs_ready($scan_mode);
      }
   }

   # 1. To restore host configuration
   &host_unconfig;

   # 2. To stop UML
   &halt;

   # 3. To remove external interfaces
   &external_if_remove;

   # 4. To remove TUN/TAPs devices
   &tun_destroy;

   # 5. To remove bridges
   &bridges_destroy;

   # To remove lock file (it exists while topology is running)
   &execute ($bp{"rm"} . " $vnuml_dir/$simname/lock");
}

######################################################
# To restore host configuration

sub host_unconfig {

   # If host <host> is not present, there is nothing to unconfigure
   return if ($doc->getElementsByTagName("host")->getLength eq 0);

   # To get <host> tag
   my $host = $doc->getElementsByTagName("host")->item(0);

   # To get host interfaces list
   # never used in current version?
   #my $if_list = $host->getElementsByTagName("hostif");

   # To get host routes list
   my $route_list = $host->getElementsByTagName("route");
   for ( my $i = 0; $i < $route_list->getLength; $i++) {
       my $route_dest = &text_tag($route_list->item($i));;
       my $route_gw = $route_list->item($i)->getAttribute("gw");
       my $route_inet = $route_list->item($i)->getAttribute("type");
       # Routes for IPv4 (inet)
       if ($route_inet eq "inet") {
          if ($route_dest eq "default") {
             &execute($bp{"route"} . " -A inet del $route_dest gw $route_gw");
          } 
          else {
             &execute($bp{"route"} . " -A inet del -net $route_dest gw $route_gw");
          }
          #&execute($bp{"route"} . " -A $route_inet del $route_dest gw $route_gw");
       }
       # Routes for IPv6 (inet6)
       else {
          if ($route_dest eq "default") {
             &execute($bp{"route"} . " -A inet6 del 2000::/3 gw $route_gw");
          }
          else {
             &execute($bp{"route"} . " -A inet6 del $route_dest gw $route_gw");
          }
       }
   }

}

######################################################
# 3. To stop UML (politely) 

sub halt {

   # UMLs counter (used to generate IPv4 management addresses)
   my $counter = 0;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # Depending of how parser has been called (-F switch), the stop process is
      # more or less "aggressive" (and quick)
      # uml_mconsole is very verbose: redirect its error output to null device
      if ($opt_F) {
         # Halt trought mconsole socket
         my $mconsole = "$vnuml_dir/$simname/$name/mconsole";
         &execute($bp{"uml_mconsole"} . " $mconsole halt 2>/dev/null");
      }
      else {
         # Halt command in the UML
	 my ($net, $mask, $host1, $host2) = get_admin_address($counter);
	 &execute($bp{"ssh"} . " -x -1 -l root $host2 halt");

      }
      $counter++;
   }

   # For no-forced mode, we have to wait all UMLs dead before to destroy 
   # TUN/TAP (next step) due to these devices are yet in use
   #
   # Note that -B doensn't affect to this functionallity. UML extinction
   # blocks can't be avoided (is needed to perform bridges and interfaces
   # release)
   my $time_0 = time(); 
   if ((!$opt_F)&&($exe_mode != EXE_DEBUG)) {
   print "---------- Waiting until UML extinction ----------\n" if ($exe_mode == EXE_VERBOSE);
   while (&UML_alive) {
         #system($bp{"sleep"} . " $delay");
         sleep($delay);
	 my $time_f = time();
         my $interval = $time_f - $time_0;
         print "$interval seconds eslapsed...\n" if ($exe_mode == EXE_VERBOSE);
      }
   }
}


######################################################
# To remove external interfaces

sub external_if_remove {

   # To get list of defined <net>
   my $net_list = $doc->getElementsByTagName("net");

   # To process list, decreasing use counter of external interfaces
   for ( my $i = 0; $i < $net_list->getLength; $i++ ) {
      my $net = $net_list->item($i);

      # To get name attribute
      my $name = $net->getAttribute("name");

      # We check if there is an associated external interface
      my $external_if = $net->getAttribute("external");
      next if ($external_if =~ /^$/);

      # To check if VLAN is being used
      my $vlan = $net->getAttribute("vlan");
      $external_if .= ".$vlan" unless ($vlan =~ /^$/);

      # To decrease use counter
      &dec_if($external_if);

      # To clean up not in use physical interfaces
      if (&get_if($external_if) == 0) {
         &execute($bp{"ifconfig"} . " $name 0.0.0.0 $promisc up");
         &execute($bp{"brctl"} . " delif $name $external_if");
	 unless ($vlan =~ /^$/) {
	    &execute($bp{"vconfig"} . " rem $external_if");
	 }
	 else {
	    # Note that now the interface has no IP address nor mask assigned, it is
            # unconfigured! Tag <physicalif> is checked to try restore the interface
            # configuration (if it exists)
	    &physicalif_config($external_if);
	 }
      }
   }

}

######################################################
# To remove TUN/TAPs devices

sub tun_destroy {

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # To throw away and remove management device (id 0)
      my $tun_if = $name . "-eth0";
      &execute($bp{"ifconfig"} . " $tun_if down");
      &execute($bp{"tunctl"} . " -d $tun_if -f $tun_device");

      # To get UML's interfaces list
      my $if_list = $vm->getElementsByTagName("if");

      # To process list
      for ( my $j = 0; $j < $if_list->getLength; $j++) {
         my $if = $if_list->item($j);

         # To get id attribute
         my $id = $if->getAttribute("id");

         # To build TUN device name
	 my $tun_if = $name . "-eth" . $id;

	 # To throw away TUN device
	 &execute($bp{"ifconfig"} . " $tun_if down");

         # To remove TUN device
	 &execute($bp{"tunctl"} . " -d $tun_if -f $tun_device");

      }

   }

}

######################################################
# To remove bridges

sub bridges_destroy {

   # To get list of definde <net>
   my $net_list = $doc->getElementsByTagName("net");

   # To process list, decreasing use counter of external interfaces
   for ( my $i = 0; $i < $net_list->getLength; $i++ ) {
      my $net = $net_list->item($i);

      # To get name attribute
      my $name = $net->getAttribute("name");

      # Set bridge down and remove it only in the case there isn't any associated interface 
      if (&vnet_ifs($name) == 0) {
         &execute($bp{"ifconfig"} . " $name down");
         &execute($bp{"brctl"} . " delbr $name");
      }

   }

}

#####################################################

# mode_P
#
# Purge simulation
#
# POTENCIAL PROBLEM: what's happen if the <kernel> has a name diffeent from 'linux'?
sub mode_P {
   
   my $pids;

   # 1. Kill all Linux processes, gracefully
   $pids = join(" ",&get_kernel_pids);
   &execute($bp{"kill"} . " -SIGTERM $pids"); 
   print "Waiting UMLs to term gracefully...\n" unless ($exe_mode == EXE_NORMAL);
   sleep ($delay) unless ($exe_mode == EXE_DEBUG);

   # 2. Kill all remaining Linux processes, by brute force
   $pids = join(" ",&get_kernel_pids);
   &execute($bp{"kill"} . " -SIGKILL $pids");
   print "Waiting remaining UMLs to term forcely...\n" unless ($exe_mode == EXE_NORMAL);
   sleep ($delay) unless ($exe_mode == EXE_DEBUG);

   # 3. Delete supporting simulation files
   &execute ($bp{"rm"} . " -r $vnuml_dir/$simname/*");

}


######################################################

# Additional functions

# execute
#
# Shell commands execution interface. This functions is
# based on the value of global variable $exe_mode
#
# Depending of the number of arguments, execute works in direct
# or or recording mode (execpt for EXE_DEBUG).
#
# * Direct mode: command execution (first argument) throught system
#   Perl function.
# * Recording mode: command execution (first argument) printing it
#   to the flow pointed by the second argument. This function is for
#   dumping commands to a file, that will be executed in another momment (for
#   example, UML boot script)
#
# Interactive mode can be used in addition to direct mode ($exe_interactive)
#
# Verbose output (when $exe_mode has EXE_DEBUG or EXE_VERBOSE)
# goes to standard output. It is prefixed with $verb_prompt
# and, in addition, whit D- in debug mode

sub execute {

   my $retval = 0;	# By default, all right
   if ((my ($command, $CMD_OUT) = @_) == 1) {
      # Direct mode
      if ($exe_mode == EXE_DEBUG) {
         print "D-" . $verb_prompt . "$command\n";
      }
      elsif ($exe_mode == EXE_VERBOSE) {
         print $verb_prompt . "$command\n";
         system $command;
         $retval = $?;
	 if ($exe_interactive) {
            &pulse_a_key;
	 }
      }
      elsif ($exe_mode == EXE_NORMAL) {
         system $command;
         $retval = $?;
	 if ($exe_interactive) {
            &pulse_a_key;
	 }
      }
   }
   else {
      # Recording mode
      if ($exe_mode == EXE_DEBUG) {
         print "D-" . $verb_prompt . "$command\n";
      }
      elsif ($exe_mode == EXE_VERBOSE) {
         print $verb_prompt . "$command\n";
         print $CMD_OUT "$command\n";
      }
      elsif ($exe_mode == EXE_NORMAL) {
         print $CMD_OUT "$command\n";
      }
   }

   return $retval;

}

sub pulse_a_key {

   # Copy-paste from http://perlmonks.thepen.com/33566.html
   # A simpler alternative to this code is <>, but that is ugly :)

   use Term::ReadKey;
   my $key;
   ReadMode 4; # Turn off controls keys
   while (not defined ($key = ReadKey(1))) {
      # No key yet
   }
   ReadMode 0; # Reset tty mode before exiting

}

# show_file	NOT USED AT CURRENT VERSION
#
# Shows file specified with first argument.
# Every line is prefixed with second argument as prompt
#sub show_file {
#   open SHOWFILE, "$_[0]";
#   while (<SHOWFILE>) {
#      print $_[1] . $_;
#   }
#   close SHOWFILE;
#}

# check_binaries
#
# Check if the binaries needed for VNUML operation are available.
# First argument is the list of binaries to check. Return the
# number of unchecked binaries (0 if all the binaries were found).
#
# Path of checked binaries are stored in %bp array for use during
# VNUML operation
#
sub check_binaries {

   my $unchecked = 0;
   foreach (@_) {
      print "Checking $_... " if (($exe_mode == EXE_VERBOSE) || ($exe_mode == EXE_DEBUG));
      my $fail = system("which $_ &> /dev/null");
      if ($fail) {
         print " not found\n" if (($exe_mode == EXE_VERBOSE) || ($exe_mode == EXE_DEBUG));;
	 $unchecked++;
      }
      else {
         my $where = `which $_`;
         chomp($where);
         print "$where\n" if (($exe_mode == EXE_VERBOSE) || ($exe_mode == EXE_DEBUG));;
	 $bp{$_} = $where;
      }
   }
   return $unchecked; 

}

# check_semantics
#
# Checks additional semantics in VNUML file that can not be
# checked during XML validation process.
# Currently this check consist in:
#
#   - to control source file version
#   - there is not duplicated net names (<net>)
#   - external interfaces (external attribute of <net>) are actually physical interfaces
#     (they exist in the OS host enviroment)
#   - there is not duplicated UML names (<vm>)
#   - there is not duplicated interfaces in the same UML (<if>)
#   - there is not <if> with id 0 in any UML (reserved for management)
#   - net attribute of <if> and <hostif> refers to existing <net>
#   - there is not duplicated <physicalif> names (we would be defining twice the same thing)
#     and they exists in the OS host enviroment
#   - <vm> and <net> names max of $max_name_length characters
#   - <ipv4> valid addresses
#   - <ipv6> valid addresses
#   - check IPv4 or IPv6 addresses for destination and gateway in <route> tag
#   - check ip attribute of <physicalif> for a valid IPv4 address 
#   - check IPv5 mask in <ipv4> and <physicalif> mask attributes
# 
# Check valid filename:
#
# - <ssh_key> value
# - <shell> value
# - <tun_device> value 
# - <filesystem> value
# - <kernel> value
# - <filetree> value (as directory) and "root" attribute (as directory)
# 
# Check file existence:
#
# - <ssh_key> value
# - <filesystem> value
# - <kernel> value
# - <filetree> value (as directory)
#
# Check executablility of files: 
# 
# - <kernel> value 
#
# The functions returns error string describing problem or 1 if all is right
sub check_semantics {

   # 1. Version control
   my $version_list = $doc->getElementsByTagName("version");
   if ($version_list->getLength == 0) {
      return "version of source file is missing";
   }
   my $version_tag = &text_tag($version_list->item(0));

   # Since version 1.2.1 only mayor version numbers (i.e, 1.2) are compared
   $version =~ /^(\d\.\d+)/;
   my $version_mayor = $1;
   $version_tag =~ /^(\d\.\d+)/;
   my $version_tag_mayor = $1; 

   unless ($version_tag_mayor eq $version_mayor) {
      return "mayor version numbers of source file ($version_tag) and parser ($version) do not match";
   }

   # 2. To check <ssh_key>
   my $ssh_key_list = $doc->getElementsByTagName("ssh_key");
   if ($ssh_key_list->getLength != 0) {
      my $ssh_key = &text_tag($ssh_key_list->item(0));
      return "$ssh_key is not a valid absolute filename" unless &valid_filename($ssh_key);
      return "$ssh_key is not a valid file (perhaps does not exists)" unless (-f $ssh_key);
   }

   # 3. To check <shell>
   my $shell_list = $doc->getElementsByTagName("shell");
   if ($shell_list->getLength != 0) {
      my $shell = &text_tag($shell_list->item(0));
      return "$shell is not a valid absolute filename" unless &valid_filename($shell);
   }

   # 4. To check <tun_device>
   my $tun_device_list = $doc->getElementsByTagName("tun_device");
   if ($tun_device_list->getLength != 0) {
      my $tun_device = &text_tag($tun_device_list->item(0));
      return "$tun_device is not a valid absolute filename" unless &valid_filename($tun_device);
   }

   # 5. To check <net>
   # Hash for duplicated names detection
   my %net_names;

   # To get list of defined <net>
   my $net_list = $doc->getElementsByTagName("net");

   # To process list
   for ( my $i = 0; $i < $net_list->getLength; $i++ ) {
      my $net = $net_list->item($i);

      # To get name attribute
      my $name = $net->getAttribute("name");

      # To check name length
      my $upper = $max_name_length + 1;
      return "net name $name is too long: max $max_name_length characters"
         if ($name =~ /^.{$upper,}$/);

      # To check if the same name has been seen before 
      if ($net_names{$name} == 1) {
         return "duplicated net name: $name";
      }
      else {
         $net_names{$name} = 1;
      }

      # To check external attribute (if present)
      # (this method is not very strong, only existence is checked; a best checking would
      # bet to analyze the full information returned by ifconfig command) 
      my $external_if = $net->getAttribute("external");
      unless ($external_if =~ /^$/) {
	if (system($bp{"ifconfig"} . " $external_if &> /dev/null")) {
	   return "in network $name, $external_if does not exists";
	}
      }

   }

   # 6. To check <vm> and <if>
   # Hash for duplicated names detection
   my %vm_names;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # To check name length
      my $upper = $max_name_length + 1;
      return "vm name $name is too long: max $max_name_length characters"
         if ($name =~ /^.{$upper,}$/);

      # To check if the same name has been seen before
      if ($vm_names{$name} == 1) {
         return "duplicated vm name: $name";
      }
      else {
         $vm_names{$name} = 1;
      }

      # Hash for duplicated ids detection
      my %if_ids;

      # To get UML's interfaces list
      my $if_list = $vm->getElementsByTagName("if");

      # To process list
      for ( my $j = 0; $j < $if_list->getLength; $j++) {
         my $if = $if_list->item($j);

         # To get id attribute
         my $id = $if->getAttribute("id");

	 # To check id 0 is not used
	 return "if id 0 in vm $name is not user allowed" if ($id == 0);

         # To check if the same id has been seen before
         if ($if_ids{$id} == 1) {
             return "duplicated if id $id inside vm $name";
         }
         else {
            $if_ids{$id} = 1;
         }

         # To get net attribute
         my $net = $if->getAttribute("net");

	 # To check that there is a net with this name
	 unless ($net_names{$net} == 1) {
	    return "if id $id in vm $name net $net is not valid: it must be defined in a <net> tag";
	 }

      }

      #7. To check <filesystem>
      my $filesystem_list = $vm->getElementsByTagName("filesystem");
      my $filesystem = &text_tag($filesystem_list->item(0));
      return "$filesystem is not a valid absolute filename" unless &valid_filename($filesystem);
      return "$filesystem is not a valid file (perhaps does not exists)" unless (-f $filesystem);

      #8. To check <kernel>
      my $kernel_list = $vm->getElementsByTagName("kernel");
      if ( $kernel_list->getLength != 0 ) {
         my $kernel = &text_tag($kernel_list->item(0));
         return "$kernel is not a valid absolute filename" unless &valid_filename($kernel);
         return "$kernel is not a valid file (perhaps does not exists)" unless (-f $kernel);
         return "$kernel is not an executable file (perhaps does not exists)" unless (-X $kernel);
         # Store the kernel name
         $kernel_names{$kernel} = 1;
      }
      else  { 
         return "default kernel /usr/local/bin/linux is not a valid file (perhaps does not exists)" unless (-f "/usr/local/bin/linux");
         return "default kernel /usr/local/bin/linux is not an executable file" unless (-X "/usr/local/bin/linux");
         # Store the kernel name
         $kernel_names{"/usr/local/bin/linux"} = 1;
      }

      #9. To check <filetree>
      my $filetree_list = $vm->getElementsByTagName("filetree");

      # To process list
      for ( my $j = 0; $j < $filetree_list->getLength; $j++) {
         my $filetree = &text_tag($filetree_list->item($j));
         my $root = $filetree_list->item($j)->getAttribute("root");
         return "$filetree is not a valid absolute directory name" unless &valid_directoryname($filetree);
         return "$filetree is not a valid directory (perhaps does not exists)" unless (-d $filetree);
         return "$root is not a valid absolute directory name" unless &valid_directoryname($root);
      }

   }

   # 10. To check <hostif>
   my $hostif_list = $doc->getElementsByTagName("hostif");

   # To process list
   for ( my $i = 0; $i < $hostif_list->getLength; $i++) {
      my $hostif = $hostif_list->item($i);

      # To get net attribute
      my $net = $hostif->getAttribute("net");

      # To check that there is a net with this name
      unless ($net_names{$net} == 1) {
	    return "hostif net $net is not valid: it must be defined in a <net> tag";
      }

   }

   # 11. To check <physicalif>
   # Hash for duplicated names detection
   my %phyif_names;

   # To get list of defined <net>
   my $phyif_list = $doc->getElementsByTagName("physicalif");

   # To process list
   for ( my $i = 0; $i < $phyif_list->getLength; $i++ ) {
      my $phyif = $phyif_list->item($i);

      # To get name attribute
      my $name = $phyif->getAttribute("name");

      # To check if the same name has been seen before
      if ($phyif_names{$name} == 1) {
         return "duplicated physicalif name: $name";
      }
      else {
         $phyif_names{$name} = 1;
      }

      # It exists?
      if (system($bp{"ifconfig"} . " $name &> /dev/null")) {
	   return "physicalif $name does not exists";
      }

      # To get ip attribute
      my $ip = $phyif->getAttribute("ip");
      
      # To check if valid IPv4 address
      unless (&valid_ipv4($ip)) {
         return "'$ip' is not a valid IPv4 address in a <physicalif> ip";
      }

      # To get mask attribute
      my $mask = $phyif->getAttribute("mask");

      # To check if valid IPv4 mask
      unless (&valid_mask($mask)) {
         return "'$mask' is not a valid IPv4 netmask in a <physicalif> mask attribute";
      }

   }

   # 12. To check IPv4 addresses
   my $ipv4_list = $doc->getElementsByTagName("ipv4");
   for ( my $i = 0; $i < $ipv4_list->getLength; $i++ ) {
      my $ipv4 = &text_tag($ipv4_list->item($i));
      unless (&valid_ipv4($ipv4)) {
         return "'$ipv4' is not a valid IPv4 address (Z.Z.Z.Z) in a <ipv4>";
      }
      my $mask = $ipv4_list->item($i)->getAttribute("mask");
      unless (($mask eq '')||(&valid_mask($mask))) {
         return "'$mask' is not a valid IPv4 netmask in a <ipv4> mask attribute";
      }
   }

   # 13. To check IPv6 addresses
   my $ipv6_list = $doc->getElementsByTagName("ipv6");
   for ( my $i = 0; $i < $ipv6_list->getLength; $i++ ) {
      my $ipv6 = &text_tag($ipv6_list->item($i));
      unless (&valid_ipv6_with_mask($ipv6)) {
         return "'$ipv6' is not a valid IPv6 address (only Z:Z:Z:Z:Z:Z:Z:Z/M form is supported by the time) in a <ipv6>";
      }
   }

   # 14. To check addresses related with <route> tag
   my $route_list = $doc->getElementsByTagName("route");
   for ( my $i = 0; $i < $route_list->getLength; $i++ ) {
      my $route_dest = &text_tag($route_list->item($i));
      my $route_gw = $route_list->item($i)->getAttribute("gw");
      my $route_inet = $route_list->item($i)->getAttribute("type");
      if ($route_inet eq "inet") {
         unless (($route_dest eq "default")||(&valid_ipv4_with_mask($route_dest))) {
            return "'$route_dest' is not a valid IPv4 address with mask (Z.Z.Z.Z/M) in a <route>";
         }
         unless (&valid_ipv4($route_gw)) {
            return "'$route_gw' is not a valid IPv4 address (Z.Z.Z.Z) for a <route> gw";
         }
      }
      elsif ($route_inet eq "inet6") {
         unless (($route_dest eq "default")||(&valid_ipv6_with_mask($route_dest))) {
            return "'$route_dest' is not a valid IPv6 address with mask (only Z:Z:Z:Z:Z:Z:Z:Z/M is supported by the time) in a <route>";
         }
         unless (&valid_ipv6($route_gw)) {
            return "'$route_gw' is not a valid IPv6 address (only Z:Z:Z:Z:Z:Z:Z:Z is supported by the time) in a <route> gw";
         } 
      }
      else {
         return "$route_inet is not a valid <route> type";
      }

   }

   return 0;

}

# valid_filename
#
# Check if the argument (string) if a valid absolute filename
#
# Return 1 if valid, 0 otherwise
#
sub valid_filename {

   my $test_name = shift;
   
   if ($test_name =~ /^\/((\w|\.|-)+\/)*(\w|\.|-)+$/) {
      return 1;
   }
   else {
      return 0;
   }    

}

# valid_directoryname
#
# Check if the argument (string) if a valid absolute directory name
#
# Return 1 if valid, 0 otherwise
#
# Note that a valid_directoryname is always a valid_filename
#
sub valid_directoryname {

   my $test_name = shift;

   if ($test_name =~ /^\/((\w|\.|-)+\/)*((\w|\.|-)+)?$/) {
      return 1;
   }
   else {
      return 0;
   }

}

# valid_ipv4
#
# Check if the argument (string) is a valid representation of IPv4 address.
#
# Return 1 if valid, 0 otherwise
#
sub valid_ipv4 {
   my $ip = shift;

   if ($ip =~ /^(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\.(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\.(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\.(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)$/) {
      return 1;
   }
   else {
      return 0;
   }
}

# valid_ipv4_with_mask
#
# Check if the argument (string) is a valid representation of IPv4 address, with a mask (after /) suffix.
#
# Return 1 if valid, 0 otherwise
#
sub valid_ipv4_with_mask {
   my $ip = shift;

   if ($ip =~ /^(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\.(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\.(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\.(25[0-5]|2[0-4]\d|1\d\d|\d\d|\d)\/(3[0-2]|[1-2]\d|\d)$/) {
      return 1;
   }
   else {
      return 0;
   }
}

# valid_mask
#
# Check if the argument (string) is a valid representation of a IPv4 mask, in dot notation
#
# For example "255.255.255.192" is a well-formed mask, "255.0.255.192" isn't
#
# Return 1 if valid, 0 otherwise
#
sub valid_mask {

   my $mask = shift;

   # This function is based converting the mask to a binary string and check if the ones (1) are
   # all in a row. OK, it's dirty but it works :) Anybody can suggest a smarter solution? 

   # Check if valid in dot notation
   unless (&valid_ipv4($mask)) {
      return 0;
   }

   # Convert to binary
   $mask =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
   my $binary_string = &dec2bin($1) . &dec2bin($2) . &dec2bin($3) . &dec2bin($4);

   # Check binary string
   unless ($binary_string =~ /^1+0*$/) {
      return 0;
   }

   return 1;

}

# valid_ipv6
#
# Check if the argument (string) is a valid representation of IPv6 address.
#
# Return 1 if valid, 0 otherwise
#
sub valid_ipv6 {
   my $ip = shift;

   if (Net::IPv6Addr::ipv6_chkip($ip)) {
      return 1;
   }
   else {
      return 0;
   }

}

# valid_ipv6_with_mask
#
# Check if the argument (string) is a valid representation of IPv6 address, with mask (after /) suffix
#
# Return 1 if valid, 0 otherwise
#
sub valid_ipv6_with_mask {

   my $addr = shift;

   # split IP and mask
   unless ($addr =~ /^(.*)\/(12[0-8]|1[0-1]\d|\d\d|\d)$/) {
      return 0;
   }

   # check IP
   my $ip = $1;

   if (Net::IPv6Addr::ipv6_chkip($ip)) {
      return 1;
   }
   else {
      return 0;
   }

}

# dec2bin and bin2dec adapted from http://perlmonks.thepen.com/2664.html
#
sub dec2bin {
   my $str = unpack("B32", pack("N", shift));
   #$str =~ s/^0+(?=\d)//;   # otherwise you'll get leading zeros
   $str =~ s/^0*(\d{8})$/$1/;	# only the last 8 bits
   return $str;
}
sub bin2dec {
   return unpack("N", pack("B32", substr("0" x 32 . shift, -32)));
}

# text_tag
#
# Gets text of the Element object that is passed as first argument.
#
# Warning: this functions is a MESS, that uses DOM in a very relaxed way.
# I would like to be more smart, but if it works there is no reason to
# change, isn't it? :)
#
sub text_tag {
   my $element = $_[0];
   my @element_children = $element->getChildNodes;
   return $element_children[0]->getNodeValue;
}

# check_tag_attribute
#
# Return the number of occurrences of the attribute in the second argument
# in the tag of the first argument. Note that 0 is returned if no match
# is found
#
sub check_tag_attribute {

   my $tag = shift;
   my $attribute = shift;
   my $matches = 0;

   # To get list of defined tags
   my $tag_list = $doc->getElementsByTagName($tag);

   # To process list
   for ( my $i = 0; $i < $tag_list->getLength; $i++ ) {
      my $tag = $tag_list->item($i);

      # To try getting attribute
      my $attribute_value = $tag->getAttribute($attribute);

      unless ($attribute_value =~ /^$/) {
         $matches++;
      }

   }

   return $matches;

}

# get_admin_address
#
# Returns a four elements list:
#
# - network address
# - network mask
# - IPv4 address of one peer
# - IPv4 address of the other peer
#
# This functions uses one argument and with it calculate a TWO host network, whose
# addresses and mask returns.
#
# Private addresses of 192.168. prefix are used. Therefore, there are 2^14 maximum
# two-elements networks. This functions used one argument (an integer) to chose
# what network of this range generate.
# 
# This functions is very useful to generate addresses to being used
# for the UML management interfaces.

sub get_admin_address {

   # We take the argument as a 14 bits number
   my $seed = shift;

   # To offset the number
   $seed += $ip_offset;

   # The mask is always /31, the longest possible for two-host networks
   my $mask = "255.255.255.252";

   # the first two bytes are hardwired
   my $byte1 = "192";
   my $byte2 = "168";

   # the third are the most significat 8 bits of the seed
   # (the 6 bits offset to the right is to remove the less significant bits)
   my $byte3 = $seed >> 6;

   # the fourth byte are the 6 less significant bytes of seed
   # filled with 2 zeros to the right
   my $byte4 = ($seed << 2 ) & 63;

   # To (weakly) check range (all bytes in 0-255 range)
   if (($byte3 > 255) || ($byte3 < 0) || ($byte4 > 252) || ($byte4 < 0)) {
      &smartdie ("IPv4 address out of range at get_admin_address. Wrong <ip_offset> value?\n");
   }

   # To build return addresses
   my $net = $byte1 . "." . $byte2 . "." . $byte3 . "." . $byte4;
   my $host1 = $byte1 . "." . $byte2 . "." . $byte3 . "." . ($byte4 + 1);
   my $host2 = $byte1 . "." . $byte2 . "." . $byte3 . "." . ($byte4 + 2);

   return ($net, $mask, $host1, $host2);

}

# get_kernel_pids;
#
# Return a list with the list of PID of UML kernel processes
#
sub get_kernel_pids {

   my @pid_list;
   foreach (keys %kernel_names) {
      my $partial_list = `$bp{"pidof"} $_`;
      chomp($partial_list);
      @pid_list = (@pid_list, $partial_list);
   }

   return @pid_list;

}

# UML_alive
#
# Returns 1 if there is a running UML in the process space
# of the operating system, 0 in the other case.
# Is based in a greped ps (doing the same with a pidof is not
# possible since version 1.2.0)
#
# This functions is similar to UMLs_ready function
sub UML_alive {

   # To get a greped ps
   my @greped_ps;
   my $line = 0;
   #my $pipe = $bp{"ps"} . " axw | " . $bp{"grep"} . " linux |";
   my $pipe = $bp{"ps"} . " axw |";  ## changed to support <kernel> tag
   open PS_LIST, "$pipe";
   while (<PS_LIST>) {
      chomp;
      $greped_ps[$line++] = $_;
   }
   close PS_LIST;

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # To look for the current UML (if found, at least this UML is alive)
      # The name is looked for between [ ] (looking for the name alone crashes when 
      # the name of one vm could be a substring of another, like with "uml1" and "uml1b")
      my $success = 0;
      for ( my $j = 0; $j < $line; $j++) {
         $_ = $greped_ps[$j];
         return 1 if (/\($name\)/);
      }

   }
   # There is no UML running 
   return 0;

}

# UMLs_ready
#
# Returns 1 if the UMLs are ready to execute simulations, 
# 0 in other case.
#
# Arguments:
#
#    0, the check is based on a test socket 
#    otherwise, the check is based on a greped ps filter
#
sub UMLs_ready {

   my $mode = shift;

   # In debug mode, this check makes no sense, so always is true
   return 1 if ($exe_mode == EXE_DEBUG);

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   if (!$mode) {

      # UMLs counter (used to generate IPv4 management addresses)
      my $counter = 0;

      # To process list
      for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
         my $vm = $vm_list->item($i);

         # To get name attribute
         my $name = $vm->getAttribute("name");
         my ($net, $mask, $host1, $host2) = get_admin_address($counter);

         # To look for UML (if not found, at least this UML is not ready)
         my $success = 0;
         my $socket;
         if ($socket = IO::Socket::INET->new(Proto => "tcp", PeerAddr => "$host2", PeerPort => "22")) {
            $success = 1;
         }

         if (!$success) {
            print "$name sshd is not ready (socket style)\n" if ($exe_mode == EXE_VERBOSE);
            return 0;
         }
         else {
            print "$name sshd is ready (socket style)\n" if ($exe_mode == EXE_VERBOSE);
	    $counter++;
            close($socket);
         }

      }
      # All UMLs ready
      return 1;

   }
   else {
      # To get greped ps
      my @greped_ps;
      my $line = 0;
      #my $pipe = $bp{"ps"} . " axw | " . $bp{"grep"} . " linux | " . $bp{"grep"} . " sshd |";
      my $pipe = $bp{"ps"} . " axw | " . $bp{"grep"} . " sshd |";  ## changed to support <kernel> tag
      open PS_LIST, "$pipe";
      while (<PS_LIST>) {
         chomp;
         $greped_ps[$line++] = $_;
      }
      close PS_LIST;

      # To process list
      for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
         my $vm = $vm_list->item($i);

         # To get name attribute
         my $name = $vm->getAttribute("name");

         # To look for UML (if not found, at least this UML is not ready)
         my $success = 0;
         #for ( my $j = 0; $j < $line; $j++) {
         #   $_ = $greped_ps[$j];
         #   $success++ if (/$name/);
         #}
         foreach (@greped_ps) {
            $success++ if (/$name/);
         }

         if (!$success) {
            print "$name sshd is not ready (ps style)\n" if ($exe_mode == EXE_VERBOSE);
            return 0;
         }
         else {
            print "$name sshd is ready (ps style)\n" if ($exe_mode == EXE_VERBOSE);
         }


      }
      # All UMLs ready
      return 1;
   }
}

# make_vm_hash
#
# Builds a hash with the vm names of the simulation to use with -s, -t or -t modes. By
# defualt are UMLs defined in the VNUML source file are include, execpt when the -M switch. 
# is in use.
sub make_vm_hash {

   # To get list of defined <vm>
   my $vm_list = $doc->getElementsByTagName("vm");

   # To process list
   for ( my $i = 0; $i < $vm_list->getLength; $i++ ) {
      my $vm = $vm_list->item($i);

      # To get name attribute
      my $name = $vm->getAttribute("name");

      # If -M is in use, include only if there is in the list of the switch
      if ($opt_M) {
         my $vm_list = $opt_M;
	 if ($vm_list =~ /$name,|,$name,|,$name|^$name$/) {
	    $vm_hash{$name} = 1;
	 }
      }
      else {
      	$vm_hash{$name} = 1;
      }

   }

}

# automac
#
# Returns and automatic generated MAC address, using
# two arguments. Si <automac> is not in use, returns
# an empty string.
#
# The two argument have to be always diferent in each
# call to the functions, to guarantee uniqueness of
# the MAC in the simulation.
#
# Note that the use of this function limits a maximum
# of 255 UMLs with 255 interfaces each one (a more than
# reasonable limit, I think :)
#
# $automac_offset is used to complete MAC address
sub automac {

      my $ante_lower = shift;
      my $lower = shift;

      my $automac_list = $doc->getElementsByTagName("automac");
      # Si no se est usando el tag, devolvemos la cadena vacia
      if ($automac_list->getLength == 0) {
         return "";
      }

      return ",fe:fd:0:$automac_offset:$ante_lower:$lower";

}

# physicalif_config
#
# Tryes to configure the physical interface whose name is given
# as first argument. To do so, it uses information stored in
# <physicalif> attributes. This functions is used only with -d mode
sub physicalif_config {

   my $interface = shift;

   # To get list of defined <physicalif>
   my $phyif_list = $doc->getElementsByTagName("physicalif");

   # To process list
   for ( my $i = 0; $i < $phyif_list->getLength; $i++ ) {
      my $phyif = $phyif_list->item($i);

      my $name = $phyif->getAttribute("name");
      if ($name eq $interface) {
         my $ip = $phyif->getAttribute("ip");
	 my $mask = $phyif->getAttribute("mask");
	 my $gw = $phyif->getAttribute("gw");
	 &execute($bp{"ifconfig"} . " $interface $ip netmask $mask");
	 &execute($bp{"route"} . " add default gw $gw");
      }

   }

}

# exists_simulation
#
# Returns true if the simulation (first argument) is currently running
#
# In the current version, this check is perform looking for the lock file
# in the directory that stores simulation files
sub simulation_exists {

   my $simulation = shift;
   if ( -f "$vnuml_dir/$simulation/lock") {
   	return 1;
   }
   else {
   	return 0;
   }

}

# vnet_exists
#
# If the virtual network (implemented with a bridge) whose name is given
# as first argument exists, returns 1. In other case, returns 0.
#
# It based in `brctl show` parsing.
sub vnet_exists {

   my $vnet_name = shift;

   # To get `brctl show`
   my @brctlshow;
   my $line = 0;
   my $pipe = $bp{"brctl"} . " show |";
   open BRCTLSHOW, "$pipe";
   while (<BRCTLSHOW>) {
      chomp;
      $brctlshow[$line++] = $_;
   }
   close BRCTLSHOW;

   # To look for virtual network processing the list
   # Note that we skip the first line (due to this line is the header of
   # brctl show: bridge name, brige id, etc.)
   for ( my $i = 1; $i < $line; $i++) {
      $_ = $brctlshow[$i];
      # We are interestend only in the first and last "word" of the line
      /^(\S+)\s.*\s(\S+)$/;
      if ($1 eq $vnet_name) {
      	# If equal, the virtual network has been found
      	return 1;
      }
   }

   # If virtual network is not found:
   return 0;

}

# vnet_ifs
#
# Returns a list in which each element is one of the interfaces (TUN/TAP devices
# or host OS physical interfaces) of the virtual network given as argument.
#
# It based in `brctl show` parsing.
sub vnet_ifs {

   my $vnet_name = shift;
   my @if_list;

   # To get `brctl show`
   my @brctlshow;
   my $line = 0;
   my $pipe = $bp{"brctl"} . " show |";
   open BRCTLSHOW, "$pipe";
   while (<BRCTLSHOW>) {
      chomp;
      $brctlshow[$line++] = $_;
   }
   close BRCTLSHOW;

   # To look for virtual network processing the list
   # Note that we skip the first line (due to this line is the header of
   # brctl show: bridge name, brige id, etc.)
   for ( my $i = 1; $i < $line; $i++) {
      $_ = $brctlshow[$i];
      # We are interestend only in the first and last "word" of the line
      /^(\S+)\s.*\s(\S+)$/;
      if ($1 eq $vnet_name) {
      	# To push interface into the list
	push (@if_list,$2);

	# Internal loop (it breaks when a line not only with the interface name is found)
      	for ( my $j = $i+1; $j < $line; $j++) {
	  $_ = $brctlshow[$j];
	  if (/^(\S+)\s.*\s(\S+)$/) {
	     last;
	  }
	  # To push interface into the list
	  /.*\s(\S+)$/;
	  push (@if_list,$1);
	}

	# The end...
	last;

      }
   }

   # To return list
   return @if_list;

}

# check_vlan
#
# This function uses two arguments. First is the name of a physical
# interface. Second is a VLAN identifier.
#
# Function returns 1 if interface has the VLAN configured, 0 in other case.
#
# The `*` can be used as wildcard. Ej: check_vlan("*","5") returns 1
# if VLAN 5 is configured in at least one interface. check_vlan("eth1","*")
# returns 1 if eth1 interface has any VLAN configured.
#
# check_vlan("*","*") always returns 1. This is a dummy use of the function.
#
# It based in `cat /proc/net/vlan/config` parsing.
sub check_vlan {

   my $if_name= shift;
   my $vlan_number = shift;

   # To get `cat /proc/net/vlan/config`
   my @catconfig;
   my $line = 0;
   my $pipe = $bp{"cat"} . " /proc/net/vlan/config |";
   open CATCONFIG, "$pipe";
   while (<CATCONFIG>) {
      chomp;
      $catconfig[$line++] = $_;
   }
   close CATCONFIG;

   # To get pair interfaz-vlan
   # Note that we skip the first line, due to this is the hader of the config file
   for ( my $i = 1; $i < $line; $i++) {
      $_ = $catconfig[$i];

      # We are interested in the last two fields of the list
      /.*\s(\S+)\s.*\s(\S+)$/;

      #print "DEBUG $if_name, $vlan_number: <$1> <$2>\n";

      # The "atoms" of the checking, set to their initial values
      my $if_ok = ($if_name eq "*") ? 1 : 0;
      my $vlan_ok = ($vlan_number eq "*") ? 1 : 0;

      # To check the line
      $vlan_ok = 1 if ($vlan_number eq $1);
      $if_ok = 1 if ($if_name eq $2);

      if ($vlan_ok && $if_ok) {
         return 1;
      }

   }

   # VLAN not found in interface:
   return 0;

}

# inc_if
#
# Increases physical interface use counter, whose name is given as first
# argument. This counter is stored in a file with the name of the interface
# (for example, "eth0" or "eth1.22") in the working directory of vnumlparser.pl
# (by default, /var/vnuml). The content of this file is a number, that
# means all simulations are currently using this interface as a brigde.
#
# The inexistence of the file is an implicit 0.
sub inc_if {

   my $interface = shift;

   unless (-f "$vnuml_dir/$interface") {
      &execute($bp{"echo"} . " 1 > $vnuml_dir/$interface");
   }
   else {
      my $command = $bp{"cat"} . " $vnuml_dir/$interface";
      my $value = `$command`;
      $value++;
      &execute($bp{"echo"} . " $value > $vnuml_dir/$interface");
   }

}

# dec_if
#
# Dual function of inc_if
sub dec_if {

   my $interface = shift;

   if (-f "$vnuml_dir/$interface") {

      my $command = $bp{"cat"} . " $vnuml_dir/$interface";
      my $value = `$command`; 
      $value--;
      if ($value == 0) {
         &execute($bp{"rm"} . " $vnuml_dir/$interface");
      }
      else {
         &execute($bp{"echo"} . " $value > $vnuml_dir/$interface");
      }

   }

}

# get_if
#
# Returns counter value.
sub get_if {

   my $interface = shift;

   unless (-f "$vnuml_dir/$interface") {
      return 0;
   }
   else {
      my $command = $bp{"cat"} . " $vnuml_dir/$interface";
      my $value = `$command`;
      return $value;
   }

}

# netconfig
#
# This functions uses has argument <netconfig> tag, where
# resides configuration to apply.
sub netconfig {

   my $net_cfg = shift;

   my $stp_att = $net_cfg->getAttribute("stp");
   $stp = 1 if ( $stp_att =~ /^on$/ );
   $stp = 0 if ( $stp_att =~ /^off$/ );

   my $promisc_att = $net_cfg->getAttribute("promisc");
   $promisc = "promisc" if ( $promisc =~ /^on$/ );
   $promisc = "" if ( $promisc =~ /^off$/ );

}

# validation_fail
#
# This function is a handler that triggers XML validator parser
# when errors are detected.
sub validation_fail {
   my $code = shift;
   # To set flag
   $valid_fail = 1;
   # To print error message
   XML::Checker::print_error ($code, @_);
}

# smartdie
#
# Wrapper of die Perl function, releasing previously global lock
sub smartdie {
   my $mess = shift;

   if (-f "$vnuml_dir/LOCK") {
      &execute($bp{"rm"} . " $vnuml_dir/LOCK");
   }
   die $mess;
}

# chompslash
#
# Removes final slashes from directory names
sub chompslash {
   my $string = shift;
   $string =~ s/\/*$//;
   return $string;
}

# usage
#
# Prints program usage message
sub usage {
   print "Usage: $basename -t VNUML_file [-o prefix] [-m mount_point] [-c vnuml_dir]\n";
   print "                 [-T tmp_dir] [-k] [-i] [-S scan_mode ] [-w delay] [-B]\n"; 
   print "                 [-e screen_file] [-v] [-g]\n";
   print "       $basename -s VNUML_file [-T tmp_dir] [-M vm_list] [-k] [-i] [-S scan_mode ] [-B] [-v] [-g]\n";
   print "       $basename -p VNUML_file [-T tmp_dir] [-M vm_list] [-k] [-i] [-S scan_mode ] [-B] [-v] [-g]\n";
   print "       $basename -r VNUML_file [-T tmp_dir] [-M vm_list] [-k] [-i] [-S scan_mode ] [-B] [-v] [-g]\n";
   print "       $basename -d VNUML_file [-c vnuml_dir] [-F] [-T tmp_dir] [-k] [-i] [-S scan_mode ] [-B] [-v] [-g]\n";
   print "       $basename -P VNUML_file [-T tmp_file] [-k] [-i] [-v] [-g]\n";
   print "       $basename -V\n";
   print "Mode:\n";
   print "       -t VNUML_file, build topology (using VNUML_file) as source\n";
   print "       -s VNUML_file, start simulation (using VNUML_file) as source\n";
   print "       -p VNUML_file, stop simulation (using VNUML_file as source)\n";
   print "       -r VNUML_file, restart simulation (using VNUML_file as source)\n";
   print "       -d VNUML_file, destroy current simulation (using VNUML_file as source)\n";
   print "       -P VNUML_file, purge simulation (WARNING: it will remove UML cowed filesystems!)\n";
   print "       -V, show program version and exits.\n";
   print "Options:\n";
   print "       -k, skips XML validation in VNUML_file (warning: it is not recommended)\n";
   print "       -o prefix, dump UML boot messages output to files (using given prefix in pathname)\n";
   print "       -m mount_point, use mount_point as mount point to configure UML filesystem (default is /mnt/uml)\n";
   print "       -c vnuml_dir, vnumlparser.pl working directory (default is /var/vnuml)\n";
   print "       -F, force stopping of UMLs (warning: UML filesystems may be corrupted)\n";
   print "       -S scan_mode, the way UML readiness is scanned (default is 0, socket style)\n";
   print "       -w delay, waits delay seconds between UMLs booting (defult is wait no time)\n";
   print "       -B, blocking mode\n";
   print "       -e screen_file, make screen configuration file for pts devices\n";
   print "       -i, interactive execution (in combination with -v mode)\n";
   print "       -v, verbose mode on\n";
   print "       -g, debug mode on (overrides verbose)\n";
   print "       -T tmp_dir, temporal files directory (default is /tmp)\n";
   print "       -M vm_list, start/stop/restart simulation in vm_list UMLs (a list of names separated by ,)\n";
}
